;+

function trace_cube_pointing,tindex,tdata,itref,mask=mask,quiet=quiet, $
                             xytrace=xytrace,no_shift_guess=no_shift_guess, $
                             _EXTRA=extra

;NAME:
;     TRACE_CUBE_POINTING
;PURPOSE:
;     Fix the pointing in a TRACE dataset by aligning each image to
;     a reference image within the data cube.  The pointing of the 
;     reference image is assumed to be correct (see trace_mdi_align.pro for
;     an example of how to get the pointing of this image correct).  The 
;     pointing in the new index (.xcen and .ycen) is updated to match the 
;     pointing in the reference image, but the images themselves are not 
;     modified.
;CATEGORY:
;CALLING SEQUENCE:
;     new_trace_index = trace_cube_pointing(tindex,tdata,iref)
;INPUTS:
;     tindex = trace index
;     tdata = trace data cube
;     itref = image number in the data cube which will be the reference
;             image.  If not set, assumed to be the first image.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     xytrace = 4 element array giving the LL and UR corners of a subregion
;               to use in the alignment.  If this is not set, you will be
;               asked to define the subregion with the mouse. [x0,x1,y0,y1]
;     mask = same size as the tdata cube.  Pixels with mask=1 will NOT
;            be included in the alignment calculation.  Use this to mask
;            out blemishes, saturated areas, whatever.  If mask exists, it
;            must exist for every image in the data cube.
;     /no_shift_guess = ignore the pointing shift between the images and the
;                       reference image when computing the initial guess for
;                       the transformation between them.  Useful if the images
;                       have nearly the same pointing to start with, but the 
;                       pointing error in the images is large so that the 
;                       corrected pointing in the reference image is far from
;                       the incorrect pointing in the images to be fixed.
;     quiet = quiet
;OUTPUTS:
;     new_trace_index = new trace index with .xcen and .ycen updated to match
;                       the pointing of the reference image.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Very slow.
;
;     Solar rotation is accounted for but this will not give a correct
;     change for coronal images.
;PROCEDURE:
;     Uses auto_align_images with a second order transformation to allow
;     for a mask and for foreshortening, etc.  Only the .xcen and .ycen 
;     fields in the index are updated.  The images are not in any way shifted,
;     etc. 
;MODIFICATION HISTORY:
;     T. Metcalf 2001-Oct-17
;-

ntrace = n_elements(tindex)
if n_elements(tdata[0,0,*]) NE ntrace then $
   message,'ERROR: the number of images in the cube does not equal the ' + $
           'number of index elements.'

if n_elements(itref) LE 0 then itref=0
rimage = tdata[*,*,itref]

nxr = n_elements(rimage[*,0])
nyr = n_elements(rimage[0,*])

binr = 1L+max([nxr,nyr])/640L

REPEAT nxr = nxr - 1L UNTIL nxr MOD binr EQ 0
REPEAT nyr = nyr - 1L UNTIL nyr MOD binr EQ 0

if n_elements(xytrace) NE 4 then begin
   ; Select a smaller FOV which encompasses only the features to be
   ; aligned.  This is for speed; it'd be OK to take the whole FOV.
   junk = wdefroi(rebin(rimage(0:nxr-1,0:nyr-1),nxr/binr,nyr/binr,/sample), $
                  xvr,yvr)
   xytrace = [xvr,yvr]
endif

rimage = tdata[xytrace[0]:xytrace[1],xytrace[2]:xytrace[3],itref]
tracedate0 = anytim(tindex[itref].date_obs,/yoh)
if n_elements(mask) GT 0 then begin
   rmask = where(mask[xytrace[0]:xytrace[1],xytrace[2]:xytrace[3],itref])
endif

x0 = tindex[itref].xcen/tindex[itref].cdelt1 + $
     ((xytrace[1]+xytrace[0])/2.0 - (tindex[itref].naxis1-1.)/2.)
y0 = tindex[itref].ycen/tindex[itref].cdelt2 + $
     ((xytrace[3]+xytrace[2])/2.0 - (tindex[itref].naxis2-1.)/2.)
latlon = get_latlon(nx=xytrace[1]-xytrace[0]+1L, $
                    ny=xytrace[3]-xytrace[2]+1L, $
                    x0=x0, $
                    y0=y0, $
                    b0=get_rb0p(tracedate0,/b0angle), $
                    pang=0.0, $ ; Assumes TRACE up is north
                    pixs=[tindex[itref].cdelt1,tindex[itref].cdelt2], $
                    rad=get_rb0p(tracedate0,/radius))

new_trace_index = tindex

for i=0L,ntrace-1L do begin  
   if i NE itref then begin
      if strupcase(tindex[i].wave_len) EQ $
         strupcase(tindex[itref].wave_len) then begin
         timage = tdata[xytrace[0]:xytrace[1],xytrace[2]:xytrace[3],i]

         ; First set up the initial guess for the transformation

         nxtrace = xytrace[1]-xytrace[0]+1L
         nytrace = xytrace[3]-xytrace[2]+1L
         tracedate = anytim(tindex[i].date_obs,/yoh)
         ddays = int2secarr(tracedate0,tracedate)/86400.
         ll = xy2lonlat([tindex[i].xcen,tindex[i].ycen],tracedate)
         ll[0] = ll[0] + diff_rot(ddays,ll[1],/allen) ; solar rotation
         xy = lonlat2xy(ll,tracedate)
         ; solar rotation in pixel units
         droffset = (xy-[tindex[i].xcen,tindex[i].ycen]) / $
                    [tindex[i].cdelt1,tindex[i].cdelt2]
         ; the pointing shift
         xshift = (tindex[i].xcen-tindex[itref].xcen)*tindex[i].cdelt1
         yshift = (tindex[i].ycen-tindex[itref].ycen)*tindex[i].cdelt2
         if keyword_set(no_shift_guess) then begin
            xshift = 0.0
            yshift = 0.0
         endif
         ; the total shift due to pointing and solar rotation
         xoffset = xshift + droffset[0]
         yoffset = yshift + droffset[1]
         ; The initial guess for the transformation 
         pq = rss2pq(nxtrace,nytrace, $
                     xshift=-xoffset,yshift=-yoffset, $
                     xscale=1.0,yscale=1.0,rot12=0.0, $
                     p=ptrace,q=qtrace)
         if n_elements(mask) GT 0 then begin
            tmask = where(mask[xytrace[0]:xytrace[1],xytrace[2]:xytrace[3],i])
         endif

         ; Set up the 2nd order transformations using the initial guess

         p9trace = fltarr(3,3)
         q9trace = fltarr(3,3)
         p9trace[0,0] = reform(ptrace,2,2)
         q9trace[0,0] = reform(qtrace,2,2)

         ; Use auto_align_images so that bad areas can be masked out and so
         ; that we can allow for foreshortening, etc.
         ; Warp the reference image onto the TRACE image (seems backwards
         ; but we want the transformation of the image coords to the ref 
         ; coords)

         inew = auto_align_images(rimage,timage,p9trace,q9trace,pout,qout, $
                                  tmask=rmask,rmask=tmask,/unrestricted, $
                                  scale=1.0,quiet=quiet,_EXTRA=extra)

         ; convert center of TRACE restricted FOV to reference coordinates

         xtrace = (xytrace[1]-xytrace[0])/2.0
         ytrace = (xytrace[3]-xytrace[2])/2.0
         pq2xy,xtrace,ytrace,pout,qout,xtrace_ref,ytrace_ref

         ; What is the lat/cmd of this ref point?  Remove solar rotation
         ; here since that piece of the transformation is not a pointing shift.

         lat = bilinear(latlon.latitude,xtrace_ref,ytrace_ref)
         cmd = bilinear(latlon.cmd,xtrace_ref,ytrace_ref) - $
               diff_rot(ddays,lat,/allen)
         xy = lonlat2xy([cmd,lat],tracedate)

         new_trace_index[i].xcen = ((tindex[i].naxis1-1.)/2. - $
                                    (xtrace+xytrace[0]))*tindex[i].cdelt1 + $
                                   xy[0]
         new_trace_index[i].ycen = ((tindex[i].naxis2-1.)/2. - $
                                    (ytrace+xytrace[2]))*tindex[i].cdelt2 + $
                                   xy[1]

         if NOT keyword_set(quiet) then begin
            print,'Image ',i
            print,'Old TRACE center: ',tindex[i].xcen,tindex[i].ycen
            print,'New TRACE center: ',new_trace_index[i].xcen, $
                                       new_trace_index[i].ycen
         endif
      endif else $
         message,/info,'Skipping '+string(i)+' since it does not have ' + $
                       'the same wavelength as the reference image'
   endif
endfor

return,new_trace_index

end
